# MSWordDocumentFS.py
# No answer from eric at the below email address, so hacked without his explicit permission from:
# MSWord Document Product for Zope - MSDocument.py
# Copyright (C) 2001 ric BARROCA <eric@barroca.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# MSWordDocument.py: ZClass supporting MS-Word (6/95/97/2000) documents,
#					 translating them into HTML for preview (using MSWord 
#					 module) and indexing (using ZCatalog for example).
#
# This is MSWordDocumentFS.py: a product hacked from the above
# hacked into non zodb file storage by ross.lazarus@channing.harvard.edu nov 2002
# images required more fuss than expected - in order for them to be viewable in the html,
# an ExtImg is created for each one as an object of the MSWordDocument - derived from ExtFile

__version__= '$Version$'
__author__= '$Author: rerla $'
__date__= '$Date: 2002/11/15 03:25:12 $'

from OFS import SimpleItem
from Globals import HTMLFile
from Globals import Persistent
from Globals import MessageDialog
from OFS.PropertyManager import PropertyManager
from OFS.ObjectManager import ObjectManager
from AccessControl.Role import RoleManager
from OFS import Image
from webdav.common import rfc1123_date
from OFS.SimpleItem import Item_w__name__
from OFS.content_types import guess_content_type
from DateTime import DateTime
from OFS.Cache import Cacheable
try: from cStringIO import StringIO
except: from StringIO import StringIO
import string,types

import Acquisition, re, string, os

if os.name == 'posix':
    from MSWord import MSWordDoc
else:
    from MSWordWin import MSWordDoc
    
debug = 0

if debug:
    def wdebug(msg):
            file = open('c:\mswddebug.txt', 'a')
            file.write(str(msg) + '\n')
            file.close()
bad_chars =  ' ,;()[]{}ݟ'
good_chars = '_________AAAAAAaaaaaaCcEEEEEeeeeeIIIIiiiiNnOOOOOOooooooSssUUUUuuuuYYyyZz'
TRANSMAP = string.maketrans(bad_chars, good_chars)


##Useful method to set up the id

def cookId(id, title, file):
    "Cook id and title"
    if hasattr(file, 'filename'):
        filename = file.filename
        title = title or filename
        name = filename[max(string.rfind(filename, '/'),
                          string.rfind(filename, '\\'),
                          string.rfind(filename, ':'),
                          )+1:]
    else:
        name = id
    name = string.translate(name, TRANSMAP)
    return name,title

def cookName(name):
    ext = string.lower(name[-4:])
    if debug:
        wdebug("cookName -- name: %s -- ext: %s" % (name, ext))
    if ext == ".doc":
        if debug:
            wdebug("EXT OK")
        newname = name
    else:
        if debug:
            wdebug("EXT NOT OK")
        newname = name + ".doc"
    return newname


class MSWordDocumentFS(
    SimpleItem.Item,
    Persistent,
    Item_w__name__,
    Cacheable,
    Acquisition.Implicit,
    ObjectManager,
    PropertyManager,
    RoleManager,
    ):
    """
    MSWord Document FS
    """
    
    meta_type = "MSWordDocumentFS"
    
    manage_options = (
        (
        {'label': 'Edit',		'action': 'manage_main'},
        {'label': 'Upload',		'action': 'manage_uploadForm'},
        {'label': 'View',		'action': ''},
        )
        + PropertyManager.manage_options
        + RoleManager.manage_options
        + Item_w__name__.manage_options
        + Cacheable.manage_options 
        )
    
    # Let's define some permissions
    __ac_permissions__=(
        ('View management screens', ('manage_tabs','manage_main')),
        ('Change MSWord Document', ('manage_edit','manage_editForm', 'manage', 'manage_main', 'manage_uploadForm',
                                    'manage_edit', 'manage_upload', 'PUT', 'manage_historyCopy',
                                    'manage_beforeHistoryCopy', 'manage_afterHistoryCopy')),
        ('View', ('index_html', 'viewHTML', 'download', 'showImg', 'manage_FTPget','getPDF','getHTML')),
        )

    # Two useful properties (do you see others ?)
    _properties=(
        {'id':'title', 'type': 'string'},
        {'id':'name', 'type':'string'},
        ) 


    def __init__(self, id, file, title='', relpath='/norelpath'):
        "Initialise a new instance"
        self.__name__ = id
        self.id = id
        self.title = title
        if type(file) is not type(''):
            data = file.read()
            name, title = cookId("", "", file)
        else:
            data = file
        self.relpath = relpath
        self.doc = MSWordDoc(self.id, data, self.relpath)       
        self.update_data(data, id)
        
    manage_addForm = HTMLFile('addForm', globals())

    def manage_add(self, id, file, title='', REQUEST=None):
        "Add an MSWordDocumentFS to a folder."
        self = self.this()
        relpath = self.REQUEST['URLPATH3'] # returns site relative path
        name, title = cookId(id, title, file)
        self.name = name
        self._setObject(id, MSWordDocumentFS(id, file, title, relpath))
        if REQUEST is not None:
            REQUEST['RESPONSE'].redirect(self.absolute_url()+'/manage_main')
            
    viewHTML = HTMLFile('index', globals())

    manage_main = HTMLFile('editForm', globals())

    def getHTML(self):
        """load from file"""
        self.REQUEST.RESPONSE.setHeader('Content-Type', 'text/html')
        return self.doc.getDocHTML()

    def getpdf(self):
        """load from file"""
        self.REQUEST.RESPONSE.setHeader('Content-Type', 'application/pdf')
        return self.doc.getDocPDF()

    
    def manage_edit(self, title, REQUEST=None):
        "Edit"
        self.title = title
        return MessageDialog(
            title = "Edited",
            message = "Properties for %s changed" % self.id,
            action = "manage_main",
            )

    manage_uploadForm = HTMLFile('uploadForm', globals())

    def manage_upload(self, file, REQUEST=None):
        "Upload the MSWord Document and process it (update_data)"
        if type(file) is not type(''):
            data = file.read()
        else:
            data = file
        name = self.id
        self.update_data(data, name)
        return MessageDialog(
            title = "Uploaded",
            message = "MSWord Document data for %s uploaded" % self.id,
            action = "manage_main",
            )
            
    def update_data(self, data, name):
        "Process the MSWord Document"
        for i in self.objectIds():
            self.manage_delObjects(i)
        self.size = len(data)
        self.doc = MSWordDoc(name, data, self.relpath)
        self.doc.Convert()
        self.pdfsize = self.doc.pdfsize
        path, imgs = self.doc.getImgs()
        self.imgs = imgs
        for im in imgs: # list of filenames, without paths
            try:
                self.manage_delObjects(im)
            except:
                pass
            fpath = os.path.join(self.doc.path,im)
            f = open(fpath,'r') # should get the image file?
            self.addImg(im,f,fpath,'')
            f.close()
        self.ZCacheable_invalidate()
    
    def clean(self):
        for i in self.objectIds():
            self.manage_delObjects(i)


    def importDoc(self, data):
        if hasattr(self, 'docfile'): self.manage_delObjects('docfile')
        self.manage_addFile('docfile', data, content_type="application/msword")
        return 1

    def importImgs(self, path):
        for imgname in self.imgs:
            if hasattr(self, imgname): self.manage_delObjects(imgname)
            file = open(path + imgname, 'rb')
            data = file.read()
            file.close()
            self.manage_addImage(imgname, data)
        return 1
    
    def download(self, REQUEST, RESPONSE):
        "Send the original msword file"
        self.REQUEST.RESPONSE.setHeader('Content-Type', 'application/msword')
        self.REQUEST.RESPONSE.setHeader('Content-Length', self.doc.size)
        return self.doc.getDoc()
    
    def index_html(self, REQUEST=None, RESPONSE=None):
        "Send the MSWord Document"
        if re.match('.*dav.*', REQUEST['HTTP_USER_AGENT']): 
            self.download(RESPONSE)
        else:
            RESPONSE.redirect(self.absolute_url()+'/viewHTML')
            return ''

######
## Protocol handlers
######
    def PUT(self, REQUEST, RESPONSE):
        """Handle HTTP PUT requests"""
        self.dav__init(REQUEST, RESPONSE)
        file=REQUEST['BODYFILE']
        data = file.read()
        name = cookName(self.id)
        self.update_data(data, name)        
        RESPONSE.setStatus(204)
        return RESPONSE

    #Send the MSWord document via FTP
    def manage_FTPget(self):
        self.docfile.manage_FTPget()

    #Useful method :-) for DAV and FTP clients
    def get_size(self):
        return self.size
    
    def getContentType(self):
        return "application/msword"

# below hacked from ExtFile to support images in the filesystem
# but being displayable in the html display of MSWordDocuments

    def addImg(self, id='', file='', filepath='', content_type=''):
         """ Add an image to a MSWordDocumentFS
         """
         title = id
         self = self.this()
         temp = ExtImg(id, filepath)
         self._setObject(id, temp)
         self._getOb(id).setContent(file,content_type)
     


class ExtImg(SimpleItem.Item, PropertyManager):
     """hacked from ExtFile to support images in MSWordDocumentFS
     """
     __ac_permissions__= ( 
          ('View',('index_html',
                  'icon_html',
                  'icon_gif',
                  'link',
                  'is_broken',
                  'get_size',
                  'getContentType',
                  '__str__'
                  )),
            )

     
     # MIME-Type Dictionary. To add a MIME-Type, add a file in the directory 
     # icons/_category_/_subcategory-icon-file_
     # example: Icon tifficon.gif for the MIME-Type image/tiff goes to 
     # icons/image/tifficon.gif and the dictionary must be updated like this: 
     # 'image':{'tiff':'tifficon.gif','default':'default.gif'}, ...
     _types={'image':                        
                    {'default':'default.gif'},
               'text':
                    {'html':'html.gif', 'xml':'xml.gif', 'default':'default.gif', 
                     'python':'py.gif'},
               'application':
                    {'pdf':'pdf.gif', 'zip':'zip.gif', 'tar':'zip.gif', 
                     'msword':'doc.gif', 'excel':'xls.gif', 'powerpoint':'ppt.gif', 
                     'default':'default.gif'},
               'video':
                    {'default':'default.gif'},
               'audio':
                    {'default':'default.gif'},
               'default':'default.gif'
          }
     
     ################################
     # Init method                  #
     ################################
     
     def __init__(self, id, filepath): 
          """ initialize a new instance of ExtFile """
          self.id = id
          self.filepath = filepath
          self.content_type = ''
               
     ################################
     # Public methods               #
     ################################
     
     def __str__(self): return self.index_html()
     
     def __len__(self): return 1
     
     def index_html (self, icon=0, preview=0, width=None, height=None, 
                     REQUEST=None):
          """ return the file with it's corresponding MIME-type """
          # HTTP If-Modified-Since header handling. (Copied from OFS/Image.py)
          if REQUEST is None and hasattr(self,'REQUEST'): 
               REQUEST = self.REQUEST
          if REQUEST is not None:
               header = REQUEST.get_header('If-Modified-Since', None)
               if header is not None:
                    header = string.split(header, ';')[0]
                    try:    mod_since = long(DateTime(header).timeTime())
                    except: mod_since = None
                    if mod_since is not None:
                         if self._p_mtime:
                              last_mod = long(self._p_mtime)
                         else:
                              last_mod = long(0)
                         if last_mod > 0 and last_mod < mod_since:
                              RESPONSE.setStatus(304)
                              return ''
          
          has_preview = 0
          filename = self.filepath
          content_type = self.content_type        
          cant_read_exc = "Can't read: "
          if os.path.isfile(filename):
               size = os.stat(filename)[6] # file size
          else:
               raise '%s not filename' % filename
               filename = os.path.join(SOFTWARE_HOME, 'Products',
                             'ExtFile', 'icons', 'broken.gif')
               try: size = os.stat(filename)[6]
               except: raise cant_read_exc, ("%s (%s)" %(self.id, filename))
               content_type = 'image/gif'
               icon = 1
          data = StringIO()
          self._copy(filename, data)
          if REQUEST is not None:
               REQUEST.RESPONSE.setHeader('Content-Type', content_type)
               REQUEST.RESPONSE.setHeader('Content-Length', size)
          return data.getvalue()
     
     def view_image_or_file(self):
          """ The default view of the contents of the File or Image. """
          raise 'Redirect', self.absolute_url()

     
     def is_broken(self):
          """ Check if external file exists and return true (1) or false (0) """
          fn = self.filepath
          if not isfile(fn):
                    return 1
          return 0
     
     def get_size(self):
          """ Returns the size of the file or image """
          fn = self.filename
          if isfile(fn): size = os.stat(fn)[6]
          else: size = 0
          return size
     
     rawsize = get_size
     getSize = get_size
     
     def size(self):
          """ Returns a formatted stringified version of the file size """
          return self._bytetostring(self.get_size())
     
     def getContentType(self):
          """ Get the content type of a file or image.
              Returns the content type (MIME type) of a file or image.
          """
          return self.content_type
          
     def setContent(self, file='', content_type=''):
          """ set image content type """
          self.content_type = self._get_content_type(file, file.read(100), 
                              self.id, content_type or self.content_type)

     manage_FTPget = index_html
     
     
     ################################
     # Private methods              #
     ################################
     
     def _access_permitted(self, REQUEST):
          """ check if the user is allowed to download the file """
          if hasattr(self,'use_download_permission_check') and \
             self.use_download_permission_check and \
             (REQUEST is None or not REQUEST.has_key('AUTHENTICATED_USER') or 
              not REQUEST['AUTHENTICATED_USER'].has_permission(
                                          'Download ExtFile/ExtImage', self)
             ):
               return 0
          else: 
               return 1
     
     def _get_content_type(self, file, body, id, content_type=None):
          """ determine the mime-type """
          headers = getattr(file, 'headers', None)
          if headers and headers.has_key('content-type'):
               content_type = headers['content-type']
          else:
               if type(body) is not type(''): body = body.data
               content_type, enc = guess_content_type(getattr(file,'filename',id),
                                                      body, content_type)
          return content_type
     
     def _getMIMECatAndSub(self, mime_string):
          """ Split MIME String into Category and Subcategory """
          cat = mime_string[:string.find(mime_string, '/')]   # MIME-category
          sub = mime_string[string.find(mime_string, '/')+1:] # sub-category
          return cat, sub
     
     def _copy(self, infile, outfile):
          """ read binary data from infile and write it to outfile
              infile and outfile may be strings, in which case a file with that
              name is opened, or filehandles, in which case they are accessed
              directly.
          """
          if type(infile) is types.StringType: 
               try:
                    instream = open(infile, 'rb')
               except IOError:
                    self._undo()
                    try:
                         instream = open(infile, 'rb')
                    except IOError:
                         raise IOError, ("%s (%s)" %(self.id, infile))
               close_in = 1
          else:
               instream = infile
               close_in = 0
          if type(outfile) is types.StringType: 
               try:
                    outstream = open(outfile, 'wb')
               except IOError:
                    raise IOError, ("%s (%s)" %(self.id, outfile))
               close_out = 1
          else:
               outstream = outfile
               close_out = 0
          try:
               blocksize = 2<<16
               block = instream.read(blocksize)
               outstream.write(block)
               while len(block)==blocksize:
                    block = instream.read(blocksize)
                    outstream.write(block)
          except IOError:
               raise IOError, ("%s (%s)" %(self.id, filename))
          try: instream.seek(0)
          except: pass
          if close_in: instream.close()
          if close_out: outstream.close()
     
     def _bytetostring (self, value):
          """ Convert an int-value (file-size in bytes) to an String
              with the file-size in Byte, KB or MB
          """
          bytes = float(value)
          if bytes>=1000:
               bytes = bytes/1024
               if bytes>=1000:
                    bytes = bytes/1024
                    typ = ' MB'
               else:
                    typ = ' KB'
          else:
               typ = ' Bytes'
          strg = '%4.2f'%bytes
          strg = strg[:4]
          if strg[3]=='.': strg = strg[:3]
          strg = strg+typ
          return strg
          
 
     ################################
     # Special management methods   #
     ################################
##     
##     def manage_afterClone(self, item, new_fn=None):
##          """ When a copy of the object is created (zope copy-paste-operation),
##              this function is called by CopySupport.py. A copy of the external 
##              file is created and self.filename is changed.
##          """
##          try: 
##               self.absolute_url(1) # this raises an exception, if no context
##          except: 
##               self._v_has_been_cloned=1
##          else:
##               old_fn = self._get_filename(self.filename)
##               new_fn = new_fn or self._get_new_ufn()
##               if isfile(old_fn):
##                    self._copy(old_fn, self._get_filename(new_fn))
##               self.filename = new_fn
##          return ExtFile.inheritedAttribute ("manage_afterClone") (self, item)
##          
##     def manage_afterAdd(self, item, container):
##          """ This method is called, whenever _setObject in ObjectManager gets 
##              called. This is the case after a normal add and if the object is a 
##              result of cut-paste- or rename-operation. In the first case, the
##              external files doesn't exist yet, otherwise it was renamed to .undo
##              by manage_beforeDelete before and must be restored by _undo().
##          """
##          self._undo()
##          if hasattr(self, "_v_has_been_cloned"):
##               delattr(self, "_v_has_been_cloned")
##               self.manage_afterClone(item)
##          return ExtFile.inheritedAttribute ("manage_afterAdd") \
##                 (self, item, container)
##     
##     def manage_beforeDelete(self, item, container):
##          """ This method is called, when the object is deleted. To support 
##              undo-functionality and because this happens too, when the object 
##              is moved (cut-paste) or renamed, the external file is not deleted. 
##              It are just renamed to filename.undo and remains in the 
##              repository, until it is deleted manually.
##          """
##          fn = self._get_filename(self.filename)
##          try:
##               os.rename(fn, fn+'.undo')
##          except OSError:
##               pass
##          return ExtFile.inheritedAttribute ("manage_beforeDelete") \
##                 (self, item, container)

    
"""
$Log: MSWordDocumentFS.py,v $
Revision 1.4  2002/11/15 03:25:12  rerla
Minor cosmetic cleanup

Revision 1.3  2002/11/15 03:08:39  rerla
Images now show as links which work! in the management console and show correctly in the html.

Whooohooo

Revision 1.2  2002/11/15 01:59:35  rerla
Added code to create the equivalent of an ExtFile containing each image. Not quite working (well, displaying) yet.

Revision 1.1  2002/11/14 18:41:03  rerla
First commit with new name - MSWordDocumentFS

Mostly working now

Todo:
test ftp/webdav
cleanup files in manage_after_delete
test image handling

Revision 1.8  2002/11/14 18:15:33  rerla
Mostly working now

Todo:
test ftp/webdav
cleanup files in manage_after_delete
test image handling

"""

